export ELKE27_ACCESS_CODE="12345678"
export ELKE27_PASSPHRASE="This is my passphrase"
export ELKE27_INTEGRATION=1
export ELKE27_HOST=192.168.1.xxx
export ELKE27_PORT=2101
export ELKE27_MN=0222
export ELKE27_SN=001122334455
export ELKE27_IDENTITY=0222:001122334455
export ELKE27_FWVER=1.0
export ELKE27_HWVER=1.0
export ELKE27_OSVER=1.0
export ELKE27_PIN=1234
export ELKE27_LIVE=1
export ELKE27_AREA_ID=1
export ELKE27_INTERACTIVE=true
export ELKE27_HUMAN_TEST=1
export ELKE27_ZONE_ID=31
