"""Constants used across package."""

from enum import IntEnum

REDACT_DIAGNOSTICS = True


class E27ErrorCode(IntEnum):
    ELKERR_NONE = 0

    ELKERR_UNRECOGNIZED_CMD = 11001
    ELKERR_ID_EXISTS = 11002
    ELKERR_INVALID_TABLE = 11003
    ELKERR_INVALID_PARAM = 11004
    ELKERR_NOT_ENOUGH_SPACE = 11005
    ELKERR_INVALID_ID = 11006
    ELKERR_TIMEOUT = 11007
    ELKERR_NOAUTH = 11008
    ELKERR_MISSING = 11009
    ELKERR_CALLBACK_INVALID = 11010
    ELKERR_NO_RESP = 11011
    ELKERR_DUPLICATE_PIN = 11012
    ELKERR_DUPLICATE_NAME = 11013
    ELKERR_INVALID_MEDIA = 11014
    ELKERR_NOT_READY = 11015
    ELKERR_IN_USE = 11016
    ELKERR_STORAGE_NOT_READY = 11017
    ELKERR_UNKNOWN_BUS_MODULE = 11018
    ELKERR_IN_USE_BY_RULE = 11019
    ELKERR_STRING_TRUNCATED = 11020
    ELKERR_INVALID_CACHE = 11021
    ELKERR_INVALID_MODE = 11022
    ELKERR_NOT_BYPASSABLE = 11023
    ELKERR_REBOOT_REQ = 11024
    ELKERR_CLOUD_BROKER_ALREADY_CONNECTED = 11025
    ELKERR_CLOUD_TOO_MANY_SESSIONS = 11026
    ELKERR_IN_ALARM = 11027
    ELKERR_NOT_ALLOWED_WHEN_ARMED = 11028
    ELKERR_2WSMK_NOT_ALLOWED = 11029
    ELKERR_INCOMPATIBLE = 11030
    ELKERR_UPDATING = 11031
    ELKERR_DUPLICATE_SN = 11032
    ELKERR_INVALID_AREA = 11033
    ELKERR_NOT_ENABLE = 11034
    ELKERR_RCVD_DONE = 11035
    ELKERR_INVALID_SESSION = 11036
    ELKERR_INVALID_PIN = 11037
    ELKERR_DURESS_USER_NOT_ALLOWED = 11038
    ELKERR_ZWAVE_BUSY = 11039
    ELKERR_ZWAVE_DONE = 11040

    ELKERR_UNKNOWN = 0xFFFFFFFF
